import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;
import org.xml.sax.*;
import java.io.*;
import org.w3c.dom.*;
import java.util.*;

public class WebDDBuilder {

    public static void main(String argv[]) {
        int argCount = argv.length;
        if (argCount < 2 || argCount >3) {
            System.err.println("Usage: DOMParse servlet-class URL-mapping [DD-file]");
            System.exit(1);
        }
        String servletClass = argv[0];
        String URLPattern = argv[1];
        try {
            WebDDBuilder dd = new WebDDBuilder(argCount == 3? argv[2]: null);
            dd.addServlet(servletClass, URLPattern);
            // output document
            dd.print(System.out);
        }
        catch (IllegalArgumentException ex) {
            System.err.println ("Invalid argument" + ex);
            ex.printStackTrace(System.out);
        }
    }

    private static final String SERVLET_VERSION = "2.4";
    private static final String XML_NAMESPACE = "http://java.sun.com/xml/ns/j2ee";
    private static final String XML_SCHEMA_INST = "http://www.w3.org/2001/XMLSchema-instance";
    private static final String XML_SCHEMA_LOC = "http://java.sun.com/xml/ns/j2ee http://java.sun.com/xml/ns/j2ee/web-app_2_4.xsd";

    private static final String SERVLET = "servlet";
    private static final String SERVLET_MAPPING = "servlet-mapping";
    private static final String SERVLET_NAME = "servlet-name";
    private static final String SERVLET_CLASS = "servlet-class";
    private static final String URL_PATTERN = "url-pattern";

    private static final String[] DD_ELEMENTS = {"icon", "display-name",
        "description", "distributable", "context-param", "filter",
        "filter-mapping", "listener", "servlet", "servlet-mapping",
        "session-config", "mimemapping", "welcome-file-list", "error-page",
        "taglib", "resource-env-ref", "resource-ref", "security-constraint",
        "login-config", "security-role", "env-entry", "ejb-ref", "ejb-local-ref" };

    private Document document;
    private Element root;
    private HashMap DDElements;

    public WebDDBuilder () {
        this(null);
    }

    public WebDDBuilder (String fileName) {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(false);
        try {
            DocumentBuilder builder = factory.newDocumentBuilder();
            if (fileName != null) {
                document = builder.parse(new File(fileName));
                document.getDocumentElement().normalize ();
                root = document.getDocumentElement();
            }
            else {
                document = builder.newDocument();
                root = document.createElement("web-app");
                root.setAttribute("version", SERVLET_VERSION);
                root.setAttribute("xmlns", XML_NAMESPACE);
                root.setAttribute("xmlns:xsi", XML_SCHEMA_INST);
                root.setAttribute("xsi:schemaLocation", XML_SCHEMA_LOC);
            }
            DDElements = createDDMap(DD_ELEMENTS);
        }
        catch (ParserConfigurationException ex) {
           System.err.println ("Failed to create DOM document:" + ex);
        }
        catch (IOException ex) {
            System.err.println ("Failed to open file:" + fileName + " " + ex);
        }
        catch (SAXException ex) {
            System.err.println ("General SAX exeception:" + ex);
        }
    }

    private void addServlet (String servletClass, String URLPattern) {
        // creates a DOM representation of a web xml deployment descriptor

        //create the servlet name from the servlet class name
        // if fully qualified class name take just last part
        int index = servletClass.lastIndexOf(".");
        String servletName;
        if (index != -1)
            servletName = servletClass.substring(index+1);
        else
            servletName = servletClass;

        //** new function
        // check to see if servlet already in dd
        // this will throw an exception if servlet already exists
        checkExists (servletName);

        // build the servlet element
        Element servlet_name = document.createElement(SERVLET_NAME);
        servlet_name.appendChild(document.createTextNode(servletName));

        Element servlet_class = document.createElement(SERVLET_CLASS);
        servlet_class.appendChild(document.createTextNode(servletClass));

        Element servlet = document.createElement (SERVLET);
        servlet.appendChild(servlet_name);
        servlet.appendChild(servlet_class);

        // find where in the DOM to insert the new servlet node
        Node refChild = findNode (root, DDElements, SERVLET);
        root.insertBefore(servlet, refChild);

        // build the servlet-mapping element
        Element url_pattern = document.createElement(URL_PATTERN);
        url_pattern.appendChild(document.createTextNode(URLPattern));

        Element servlet_mapping = document.createElement (SERVLET_MAPPING);
        // no need to create servlet name element as we already have one
        // make sure we clone deep so that we get the text node
        servlet_mapping.appendChild(servlet_name.cloneNode(true));
        servlet_mapping.appendChild(url_pattern);

        refChild = findNode (root, DDElements, SERVLET_MAPPING);
        root.insertBefore(servlet_mapping, refChild);
    }

    private void print (PrintStream stream) {
        try {
            TransformerFactory tf = TransformerFactory.newInstance();
            Transformer transformer = tf.newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT,"yes");
            transformer.transform(new DOMSource(root), new StreamResult(stream));
        }
        catch (TransformerConfigurationException ex) {
            System.err.println ("Failed to create transformer factory:" + ex);
        }
        catch (TransformerException ex) {
            System.err.println ("Failed to transform DOM tree:" + ex);
        }
    }

    private Node findNode (Node treeRoot, HashMap ddSchema, String tagName) {

        // find out index of tagName
        int refKey = getKey (ddSchema, tagName);

        NodeList tags = treeRoot.getChildNodes();
        int tagsLen = (tags != null) ? tags.getLength() : 0;

        // find first tag after tagName in tree
        for (int i = 0; i < tagsLen; i++) {
            Node tag = tags.item(i);
            if (getKey(ddSchema, tag.getNodeName()) > refKey)
                return tag;
        }
        return null;
    }

    private int getKey (HashMap ddSchema, String tagName) {
        for (int key = 0; key < ddSchema.size(); key++) {
            if (ddSchema.get(new Integer(key)).equals(tagName))
                return key;
        }
        return -1; 
    }

    private HashMap createDDMap(String[] ddSchema) {
        HashMap map = new HashMap();
        for (int i = 0; i < ddSchema.length; i++)
            map.put(new Integer(i), ddSchema[i]);
        return map;
    }

    private void checkExists (String tagName) {
        NodeList servlets = root.getElementsByTagName(SERVLET_NAME);
        int servletsLen = (servlets != null) ? servlets.getLength() : 0;

        for (int tag = 0; tag < servletsLen; tag++) {
            Node node = servlets.item(tag);

            // find text node
            NodeList children = node.getChildNodes();
            int childLen = (children != null) ? children.getLength() : 0;
            for (int i = 0; i < childLen; i++) {
                Node child = children.item(i);
                if (child.getNodeType() == Node.TEXT_NODE  &&
                    child.getNodeValue().equals(tagName))
                    // servlet already defined
                    throw (new java.lang.IllegalArgumentException());
            }
        }
    }
}


